# Copyright (c) 2012, DangerOnTheRanger
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met: 
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer. 
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution. 


__author__ = "DangerOnTheRanger"
__date__ = "May 12, 2012 2:30:10 AM"


import os
import time
import uuid

import panda3d.core
import panda3d.ai

import alias.utils
import alias.shadow


BULLET_SPEED = 300
RELOAD_TIME = 1.0


class Enemy(object):

    Z_OFFSET = 0
    SCALE = 1

    def __init__(self, position, heading):

        self._initial_position = position
        self._initial_heading = heading

        self._health = 100

    def load(self, window, ai_world):

        self._ai_world = ai_world
        self._window = window

        self._load_model()
        self._model.reparentTo(self._window.render)
        self._model.setPos(self._initial_position)
        self._model.setScale(self.SCALE)
        self._model.setH(self._initial_heading)
        self._model.setZ(self._model, self.Z_OFFSET)

        self._bullets = []
        self._time_since_last_bullet = 0.0
        self._window.taskMgr.add(self._bullet_movement_task, 'Bullet movement')

        self._load_ai()

    def take_damage(self, damage):

        self._health -= damage
        if self._health <= 0:
            self._destroy()

    def _load_model(self):
        raise NotImplementedError

    def _load_ai(self):
        raise NotImplementedError

    def _create_bullet(self):

        card_maker = panda3d.core.CardMaker('bullet maker')
        BULLET_SIZE = 1.5
        card_maker.setFrame(-BULLET_SIZE / 2.0, BULLET_SIZE / 2.0, -BULLET_SIZE / 2.0, BULLET_SIZE / 2.0)
        card_maker.setHasUvs(True)

        bullet_geometry = card_maker.generate()
        bullet_texture = self._window.loader.loadTexture(os.path.join(alias.utils.get_data_directory(),
                                                                      'textures',
                                                                      'enemy-shot.png'))

        bullet_node = panda3d.core.NodePath(bullet_geometry)
        bullet_node.setTexture(bullet_texture)
        bullet_node.setTransparency(panda3d.core.TransparencyAttrib.MAlpha)
        bullet_node.setBillboardPointWorld()
        bullet_node.setCollideMask(panda3d.core.BitMask32(0x0))
        dummy_bullet_node = panda3d.core.NodePath('enemy_bullet')
        bullet_node.reparentTo(dummy_bullet_node)

        collision_geometry = panda3d.core.CollisionSphere(0, 0, 0, BULLET_SIZE / 2.0)
        collision_node = dummy_bullet_node.attachNewNode(panda3d.core.CollisionNode('bullet collision node'))
        collision_node.node().addSolid(collision_geometry)
        collision_node.node().setFromCollideMask(panda3d.core.GeomNode.getDefaultCollideMask() | 0x1)
        collision_handler = panda3d.core.CollisionHandlerQueue()

        self._window.cTrav.addCollider(collision_node, collision_handler)

        return dummy_bullet_node, collision_handler, collision_node

    def _bullet_movement_task(self, task):

        for index, (bullet, bullet_collision_handler, bullet_collision_node) in list(enumerate(self._bullets)):

            if bullet_collision_handler.getNumEntries() > 0:

                bullet_collision_handler.sortEntries()
                colliding_object = bullet_collision_handler.getEntry(0).getIntoNodePath()
                colliding_object = colliding_object.findNetPythonTag('player')

                if colliding_object.isEmpty() is False:

                    player = colliding_object.getPythonTag('player')
                    player.kill()

                self._window.cTrav.removeCollider(bullet_collision_node)
                bullet.removeNode()
                del self._bullets[index]

            else:

                time_since_last_frame = globalClock.getDt()
                bullet.setY(bullet, BULLET_SPEED * time_since_last_frame)

        return task.cont

    def _destroy(self):

        self._ai_world.removeAiChar(self._name)
        self._model.detachNode()

class Sphre(Enemy):

    SCALE = 5
    Z_OFFSET = 0

    def _load_model(self):

        self._model = self._window.loader.loadModel(os.path.join(alias.utils.get_data_directory(),
                                                                 'models',
                                                                 'enemies',
                                                                 'sphre',
                                                                 'sphre'))
        self._model.setPythonTag('enemy', self)

        self._shadow = alias.shadow.make_blob_shadow(3, self._window)
        self._shadow.reparentTo(self._model)
        SHADOW_OFFSET = 3.9
        self._shadow.setZ(-SHADOW_OFFSET)

    def _load_ai(self):

        self._vision_sphere_node = self._model.attachNewNode(panda3d.core.CollisionNode('vison sphere node'))
        SPHERE_RADIUS = 30
        self._vision_sphere_node.setY(self._vision_sphere_node, 5)
        self._vision_sphere = panda3d.core.CollisionSphere(0, 0, 0, SPHERE_RADIUS)
        self._vision_sphere_node.node().addSolid(self._vision_sphere)
        self._vision_sphere_node.node().setFromCollideMask(panda3d.core.BitMask32(0x1))
        self._vision_sphere_node.setCollideMask(panda3d.core.BitMask32(0x0))
        self._vision_sphere_handler = panda3d.core.CollisionHandlerQueue()
        self._window.cTrav.addCollider(self._vision_sphere_node, self._vision_sphere_handler)
        self._window.taskMgr.add(self._vision_sphere_task, 'Vision sphere task')

        MASS = 100
        MOVEMENT_FORCE = 20
        MAX_FORCE = 50
        self._name = 'Sphre-' + str(uuid.uuid1())
        self._ai = panda3d.ai.AICharacter('Sphre-' + str(uuid.uuid1()), self._model, MASS, MOVEMENT_FORCE, MAX_FORCE)
        self._ai_world.addAiChar(self._ai)
        behaviors = self._ai.getAiBehaviors()
        behaviors.obstacleAvoidance(5.0)
        self._enable_wander()
        self._pursuing_player = False

    def _vision_sphere_task(self, task):

        if self._vision_sphere_handler.getNumEntries() > 0:

            for index in range(self._vision_sphere_handler.getNumEntries()):

                collision_entry = self._vision_sphere_handler.getEntry(index).getIntoNodePath()
                collision_entry = collision_entry.findNetPythonTag('player')

                if collision_entry.isEmpty() is False:

                    player = collision_entry

                    if self._pursuing_player is False:

                        self._pursuing_player = True

                        behaviors = self._ai.getAiBehaviors()
                        behaviors.pursue(player, 0.1)
                        behaviors.arrival(20.0)
                        behaviors.removeAi('wander')

                        self._time_since_last_bullet = time.time()

                    if time.time() >= self._time_since_last_bullet + RELOAD_TIME:

                        bullet, bullet_collision_handler, bullet_collision_node = self._create_bullet()
                        bullet.reparentTo(self._window.render)
                        bullet.setPos(self._model, 0, 5, 0)
                        bullet.lookAt(player)
                        self._bullets.append((bullet, bullet_collision_handler, bullet_collision_node))

                        self._time_since_last_bullet = time.time()


        else:
            if self._pursuing_player is True:

                self._pursuing_player = False

                self._ai.getAiBehaviors().removeAi('pursue')
                self._ai.getAiBehaviors().removeAi('arrival')
                self._enable_wander()

        return task.cont

    def _enable_wander(self):

        behaviors = self._ai.getAiBehaviors()

        WANDER_RADIUS = 10.0
        WANDER_PLANE = 0
        AREA_OF_EFFECT = 30.0
        PRIORITY = 0.5
        behaviors.wander(WANDER_RADIUS, WANDER_PLANE, AREA_OF_EFFECT, PRIORITY)
